//---------------------------------------------------------------------------
#include <vcl.h>
#include <vcl\registry.hpp>
#pragma hdrstop
#define DIRECTINPUT_VERSION 0x300
#include <dinput.h>
#include <stdio.h>
#include <dir.h>

#include "Emu92.h"
#include "Debug.h"
#define NOTYPEREDEF
#include "cpudefs.h"
#include "About.h"
#include "EmuMode.h"
#include "Entry.h"
#include "Log.h"
#include "Redef.h"

#define ROMTBL(x) readmem24_dword(readmem24_dword(0xc8)+(x)*4)
#define HANDLE_TABLE readmem24_dword(ROMTBL(0x2f)+0x11a)
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TEmu92Wnd *Emu92Wnd;

enum keys {kn,k_f1,k_f2,k_f3,k_f4,k_f5,k_f6,k_f7,k_f8,k_hand,
    k_q,k_w,k_e,k_r,k_t,k_y,k_u,k_i,k_o,k_p,k_a,k_s,k_d,k_f,
    k_g,k_h,k_j,k_k,k_l,k_z,k_x,k_c,k_v,k_b,k_n,k_m,k_shift,
    k_theta,k_dia,k_2nd,k_sto,k_space,k_equ,k_bs,k_enter1,
    k_esc,k_mode,k_clear,k_ln,k_sin,k_cos,k_tan,k_power,
    k_apps,k_enter2,k_up,k_down,k_left,k_right,k_lparan,
    k_rparan,k_comma,k_div,k_7,k_8,k_9,k_mult,k_4,k_5,k_6,
    k_minus,k_1,k_2,k_3,k_plus,k_0,k_dot,k_neg,k_on};

char* alphakeystr[]={"Q","W","E","R","T","Y","U","I","O",
    "P","A","S","D","F","G","H","J","K","L","Z","X","C",
    "V","B","N","M"};

int keymap[]={DIK_ESCAPE,k_esc,DIK_1,k_1,DIK_2,k_2,
    DIK_3,k_3,DIK_4,k_4,DIK_5,k_5,DIK_6,k_6,DIK_7,k_7,
    DIK_8,k_8,DIK_9,k_9,DIK_0,k_0,DIK_MINUS,k_neg,
    DIK_EQUALS,k_equ,DIK_BACK,k_bs,DIK_TAB,k_sto,DIK_Q,
    k_q,DIK_W,k_w,DIK_E,k_e,DIK_R,k_r,DIK_T,k_t,DIK_Y,k_y,
    DIK_U,k_u,DIK_I,k_i,DIK_O,k_o,DIK_P,k_p,DIK_LBRACKET,
    k_lparan,DIK_RBRACKET,k_rparan,DIK_RETURN,k_enter1,
    DIK_LCONTROL,k_dia,DIK_A,k_a,DIK_S,k_s,DIK_D,k_d,
    DIK_F,k_f,DIK_G,k_g,DIK_H,k_h,DIK_J,k_j,DIK_K,k_k,
    DIK_L,k_l,DIK_GRAVE,k_mode,DIK_LSHIFT,k_shift,
    DIK_BACKSLASH,k_ln,DIK_Z,k_z,DIK_X,k_x,DIK_C,k_c,
    DIK_V,k_v,DIK_B,k_b,DIK_N,k_n,DIK_M,k_m,DIK_COMMA,
    k_comma,DIK_PERIOD,k_dot,DIK_SLASH,k_div,DIK_MULTIPLY,
    k_mult,DIK_LMENU,k_2nd,DIK_SPACE,k_space,DIK_CAPITAL,
    k_hand,DIK_F1,k_f1,DIK_F2,k_f2,DIK_F3,k_f3,DIK_F4,
    k_f4,DIK_F5,k_f5,DIK_F6,k_f6,DIK_F7,k_f7,DIK_F8,
    k_f8,DIK_F9,k_apps,DIK_NUMLOCK,k_power,DIK_SCROLL,
    k_on,DIK_NUMPAD7,k_7,DIK_NUMPAD8,k_8,DIK_NUMPAD9,k_9,
    DIK_SUBTRACT,k_minus,DIK_NUMPAD4,k_4,DIK_NUMPAD5,k_5,
    DIK_NUMPAD6,k_6,DIK_ADD,k_plus,DIK_NUMPAD1,k_1,
    DIK_NUMPAD2,k_2,DIK_NUMPAD3,k_3,DIK_NUMPAD0,k_0,
    DIK_DECIMAL,k_dot,DIK_NUMPADENTER,k_enter2,DIK_DIVIDE,
    k_div,DIK_HOME,k_cos,DIK_UP,k_up,DIK_PRIOR,k_tan,
    DIK_LEFT,k_left,DIK_RIGHT,k_right,DIK_DOWN,k_down,
    DIK_INSERT,k_sin,DIK_DELETE,k_clear,DIK_SEMICOLON,
    k_theta,-1,-1};

int ti89kmat[10][8]=
    {k_hand,k_dia,k_shift,k_2nd,k_right,k_down,k_left,k_up,
     k_f5,k_clear,k_power,k_div,k_mult,k_minus,k_plus,k_enter1,
     k_f4,k_bs,k_t,k_comma,k_9,k_6,k_3,k_neg,
     k_f3,k_f6,k_z,k_rparan,k_8,k_5,k_2,k_dot,
     k_f2,k_mode,k_y,k_lparan,k_7,k_4,k_1,k_0,
     k_f1,k_cos,k_x,k_equ,k_ln,k_sin,k_sto,k_apps,
     kn,kn,kn,kn,kn,kn,kn,k_esc,
     kn,kn,kn,kn,kn,kn,kn,kn,
     kn,kn,kn,kn,kn,kn,kn,kn,
     kn,kn,kn,kn,kn,kn,kn,kn};

int ti89alphakmat[10][8]=
    {kn,kn,kn,kn,kn,kn,kn,kn,
     kn,kn,kn,k_e,k_j,k_o,k_u,kn,
     kn,kn,k_t,k_d,k_i,k_n,k_s,k_space,
     kn,kn,k_z,k_c,k_h,k_m,k_r,k_w,
     kn,kn,k_y,k_b,k_g,k_l,k_q,k_v,
     kn,kn,k_x,k_a,k_f,k_k,k_p,kn,
     kn,kn,kn,kn,kn,kn,kn,kn,
     kn,kn,kn,kn,kn,kn,kn,kn,
     kn,kn,kn,kn,kn,kn,kn,kn,
     kn,kn,kn,kn,kn,kn,kn,kn};

int ti92kmat[10][8]=
    {k_down,k_right,k_up,k_left,k_hand,k_shift,k_dia,k_2nd,
     k_3,k_2,k_1,k_f8,k_w,k_s,k_z,kn,
     k_6,k_5,k_4,k_f3,k_e,k_d,k_x,kn,
     k_9,k_8,k_7,k_f7,k_r,k_f,k_c,k_sto,
     k_comma,k_rparan,k_lparan,k_f2,k_t,k_g,k_v,k_space,
     k_tan,k_cos,k_sin,k_f6,k_y,k_h,k_b,k_div,
     k_p,k_enter2,k_ln,k_f1,k_u,k_j,k_n,k_power,
     k_mult,k_apps,k_clear,k_f5,k_i,k_k,k_m,k_equ,
     kn,k_esc,k_mode,k_plus,k_o,k_l,k_theta,k_bs,
     k_neg,k_dot,k_0,k_f4,k_q,k_a,k_enter1,k_minus};

extern void KeyDown(int,int,int,int);
extern void KeyUp(int,int,int,int);
extern void OnDown();
extern void OnUp();
extern void DoInstruction();
extern void IRQ(int n);
extern void SaveState(char *name);
extern void LoadState(char *name);

extern int calc;

void kup(int k)
{
    if (k==k_on)
    {
        OnUp();
        return;
    }
    int r92,c92,r89,c89,i,j,alpha89;
    r89=-1; c89=-1; r92=-1; c92=-1; alpha89=0;
    for (int j=0;j<10;j++)
    {
        for (int i=0;i<8;i++)
        {
            if (ti92kmat[j][i]==k)
            {
                r92=j;
                c92=i;
            }
            if (ti89kmat[j][i]==k)
            {
                r89=j;
                c89=i;
                alpha89=0;
            }
            if (ti89alphakmat[j][i]==k)
            {
                r89=j;
                c89=i;
                alpha89=1;
            }
        }
    }
    if (k==k_enter2)
    {
        r89=1;
        c89=7;
        alpha89=0;
    }
    ::KeyUp(r92,7-c92,r89,7-c89);
}

void kdown(int k)
{
    if (k==k_on)
    {
        OnDown();
        return;
    }
    int r92,c92,r89,c89,i,j,alpha89;
    r89=-1; c89=-1; r92=-1; c92=-1; alpha89=0;
    for (int j=0;j<10;j++)
    {
        for (int i=0;i<8;i++)
        {
            if (ti92kmat[j][i]==k)
            {
                r92=j;
                c92=i;
            }
            if (ti89kmat[j][i]==k)
            {
                r89=j;
                c89=i;
                alpha89=0;
            }
            if (ti89alphakmat[j][i]==k)
            {
                r89=j;
                c89=i;
                alpha89=1;
            }
        }
    }
    if (k==k_enter2)
    {
        r89=1;
        c89=7;
        alpha89=0;
    }
    if (alpha89&&(calc==89))
    {
        ::KeyDown(-1,-1,0,7);
        IRQ(1);
        for (i=0;i<256;i++)
            DoInstruction();
    }
    ::KeyDown(r92,7-c92,r89,7-c89);
    if (alpha89&&(calc==89))
    {
        IRQ(1);
        for (i=0;i<256;i++)
            DoInstruction();
        ::KeyUp(-1,-1,0,7);
        IRQ(1);
        for (i=0;i<256;i++)
            DoInstruction();
        ::KeyUp(r92,7-c92,r89,7-c89);
        IRQ(1);
        for (i=0;i<256;i++)
            DoInstruction();
    }
}

//---------------------------------------------------------------------------
int emuLink,openLink;
HANDLE hCom=NULL;
extern int debugLogLinkSend,debugLogLinkRecv;

int useInitPath=1;
extern char initPath[256];

void SendCom(int c);
int GetCom(int *c);
int CheckCom();

int clErrorFlag=0;

void clInitCom(int n,int timeout=0)
{
    char name[8],str[256];
    DCB dcb;

    sprintf(name,"COM%d:",n);
    hCom=CreateFile(name,GENERIC_READ|GENERIC_WRITE,0,NULL,
        OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
    if (!hCom)
    {
        MessageBox(NULL,"Error opening COM port","Error",MB_OK);
        return;
    }
    dcb.DCBlength=sizeof(DCB);
    dcb.BaudRate=CBR_9600; dcb.ByteSize=8; dcb.Parity=0;
    dcb.StopBits=0;
    dcb.fNull=0; dcb.fBinary=1; dcb.fParity=0;
    dcb.fOutxCtsFlow=1; dcb.fOutxDsrFlow=1;
    dcb.fDtrControl=0; dcb.fOutX=0; dcb.fInX=0;
    dcb.fErrorChar=0; dcb.fRtsControl=RTS_CONTROL_ENABLE;
    SetupComm(hCom,1024,1024);
    GetCommState(hCom,&dcb);
    sprintf(str,"COM%d: baud=9600 parity=N data=8 stop=1",n);
    BuildCommDCB(str,&dcb);
    dcb.fNull=0;
    SetCommState(hCom,&dcb);
    COMMTIMEOUTS to;
    to.ReadIntervalTimeout=MAXDWORD;
    to.ReadTotalTimeoutMultiplier=0;
    to.ReadTotalTimeoutConstant=timeout;
    to.WriteTotalTimeoutMultiplier=0;
    to.WriteTotalTimeoutConstant=0;
    SetCommTimeouts(hCom,&to);
    clErrorFlag=0;
}

void clCloseCom()
{
	if (hCom) CloseHandle(hCom);
}

void clSendCom(int ch)
{
    DWORD i;
    WriteFile(hCom,&ch,1,&i,NULL);
}

int clGetError()
{
    return clErrorFlag;
}

void clResetError()
{
    clErrorFlag=0;
}

int clGetCom(int &ch)
{
    DWORD i;
    unsigned char c;
    if (!ReadFile(hCom,&c,1,&i,NULL))
        clErrorFlag=1;
    if (!i) clErrorFlag=1;
    ch=c;
    return i;
}

extern int plusMod;
int saveStateOnExit;

extern struct
{
    char file[32];
    char name[64];
    int type;
    int version;
    int size;
} romImage[64];
extern int romImageCount;
extern int currentROM;

bool restrict=true;

__fastcall TEmu92Wnd::TEmu92Wnd(TComponent* Owner)
    : TForm(Owner)
{
    if (plusMod) Caption="Virtual TI-92+";
    emuLink=0; openLink=0;
    saveStateOnExit=1;
}
//---------------------------------------------------------------------------
extern char *ScreenLine[128];

IDirectInput *lpDI;
IDirectInputDevice *lpKeyboard;
extern HINSTANCE hInst;

void __fastcall TEmu92Wnd::Make92()
{
    if (plusMod) Caption="Virtual TI-92+";
    else Caption="Virtual TI-92";
    Application->Title=Caption;
    Image92->Show();
    Image89->Hide();
    ClientWidth=561;
    ClientHeight=306;
    LCD->Left=122;
    LCD->Top=25;
    LCD->ClientWidth=240;
    LCD->ClientHeight=128;
    KeyF1->Show(); KeyF2->Show(); KeyF3->Show();
    KeyF4->Show(); KeyF5->Show(); KeyF6->Show();
    KeyF7->Show(); KeyF8->Show(); KeyHand->Show();
    KeyA->Show(); KeyB->Show(); KeyC->Show();
    KeyD->Show(); KeyE->Show(); KeyF->Show();
    KeyG->Show(); KeyH->Show(); KeyI->Show();
    KeyJ->Show(); KeyK->Show(); KeyL->Show();
    KeyM->Show(); KeyN->Show(); KeyO->Show();
    KeyP->Show(); KeyQ->Show(); KeyR->Show();
    KeyS->Show(); KeyT->Show(); KeyU->Show();
    KeyV->Show(); KeyW->Show(); KeyX->Show();
    KeyY->Show(); KeyZ->Show(); KeyTheta->Show();
    KeyShift->Show(); KeyOn->Show(); KeyDia->Show();
    Key2ndA->Show(); KeySto->Show(); KeySpace->Show();
    KeyEqu->Show(); KeyBS->Show(); KeyEnter1->Show();
    Key2ndB->Show(); KeyEscA->Show(); KeyEscB->Show();
    KeyUp->Show(); KeyDown->Show(); KeyLeft->Show();
    KeyRight->Show(); KeyMode->Show(); KeyClear->Show();
    KeyApps->Show(); KeyEnter2->Show(); KeyLn->Show();
    KeySin->Show(); KeyCos->Show(); KeyTan->Show();
    KeyPower->Show(); KeyLParan->Show(); KeyRParan->Show();
    KeyComma->Show(); KeyDiv->Show(); KeyMult->Show();
    KeyMinus->Show(); KeyPlus->Show(); KeyEnter3->Show();
    Key0->Show(); KeyDot->Show(); KeyNeg->Show();
    Key1->Show(); Key2->Show(); Key3->Show();
    Key4->Show(); Key5->Show(); Key6->Show();
    Key7->Show(); Key8->Show(); Key9->Show();
    K89_F1->Hide(); K89_F2->Hide(); K89_F3->Hide();
    K89_F4->Hide(); K89_F5->Hide(); K89_2nd->Hide();
    K89_Shift->Hide(); K89_Esc->Hide(); K89_Up->Hide();
    K89_Down->Hide(); K89_Left->Hide(); K89_Right->Hide();
    K89_Dia->Hide(); K89_Alpha->Hide(); K89_Apps->Hide();
    K89_Home->Hide(); K89_Mode->Hide(); K89_Catalog->Hide();
    K89_BS->Hide(); K89_Clear->Hide(); K89_X->Hide();
    K89_Y->Hide(); K89_Z->Hide(); K89_T->Hide();
    K89_Power->Hide(); K89_Equ->Hide(); K89_LParan->Hide();
    K89_RParan->Hide(); K89_Comma->Hide(); K89_Div->Hide();
    K89_I->Hide(); K89_7->Hide(); K89_8->Hide();
    K89_9->Hide(); K89_Mult->Hide(); K89_EE->Hide();
    K89_4->Hide(); K89_5->Hide(); K89_6->Hide();
    K89_Minus->Hide(); K89_Sto->Hide(); K89_1->Hide();
    K89_2->Hide(); K89_3->Hide(); K89_Plus->Hide();
    K89_On->Hide(); K89_0->Hide(); K89_Dot->Hide();
    K89_Neg->Hide(); K89_Enter->Hide();
}

void __fastcall TEmu92Wnd::Make89()
{
    Caption="Virtual TI-89";
    Application->Title=Caption;
    Image92->Hide();
    Image89->Show();
    ClientWidth=201;
    ClientHeight=456;
    LCD->Left=18;
    LCD->Top=31;
    LCD->ClientWidth=160;
    LCD->ClientHeight=100;
    KeyF1->Hide(); KeyF2->Hide(); KeyF3->Hide();
    KeyF4->Hide(); KeyF5->Hide(); KeyF6->Hide();
    KeyF7->Hide(); KeyF8->Hide(); KeyHand->Hide();
    KeyA->Hide(); KeyB->Hide(); KeyC->Hide();
    KeyD->Hide(); KeyE->Hide(); KeyF->Hide();
    KeyG->Hide(); KeyH->Hide(); KeyI->Hide();
    KeyJ->Hide(); KeyK->Hide(); KeyL->Hide();
    KeyM->Hide(); KeyN->Hide(); KeyO->Hide();
    KeyP->Hide(); KeyQ->Hide(); KeyR->Hide();
    KeyS->Hide(); KeyT->Hide(); KeyU->Hide();
    KeyV->Hide(); KeyW->Hide(); KeyX->Hide();
    KeyY->Hide(); KeyZ->Hide(); KeyTheta->Hide();
    KeyShift->Hide(); KeyOn->Hide(); KeyDia->Hide();
    Key2ndA->Hide(); KeySto->Hide(); KeySpace->Hide();
    KeyEqu->Hide(); KeyBS->Hide(); KeyEnter1->Hide();
    Key2ndB->Hide(); KeyEscA->Hide(); KeyEscB->Hide();
    KeyUp->Hide(); KeyDown->Hide(); KeyLeft->Hide();
    KeyRight->Hide(); KeyMode->Hide(); KeyClear->Hide();
    KeyApps->Hide(); KeyEnter2->Hide(); KeyLn->Hide();
    KeySin->Hide(); KeyCos->Hide(); KeyTan->Hide();
    KeyPower->Hide(); KeyLParan->Hide(); KeyRParan->Hide();
    KeyComma->Hide(); KeyDiv->Hide(); KeyMult->Hide();
    KeyMinus->Hide(); KeyPlus->Hide(); KeyEnter3->Hide();
    Key0->Hide(); KeyDot->Hide(); KeyNeg->Hide();
    Key1->Hide(); Key2->Hide(); Key3->Hide();
    Key4->Hide(); Key5->Hide(); Key6->Hide();
    Key7->Hide(); Key8->Hide(); Key9->Hide();
    K89_F1->Show(); K89_F2->Show(); K89_F3->Show();
    K89_F4->Show(); K89_F5->Show(); K89_2nd->Show();
    K89_Shift->Show(); K89_Esc->Show(); K89_Up->Show();
    K89_Down->Show(); K89_Left->Show(); K89_Right->Show();
    K89_Dia->Show(); K89_Alpha->Show(); K89_Apps->Show();
    K89_Home->Show(); K89_Mode->Show(); K89_Catalog->Show();
    K89_BS->Show(); K89_Clear->Show(); K89_X->Show();
    K89_Y->Show(); K89_Z->Show(); K89_T->Show();
    K89_Power->Show(); K89_Equ->Show(); K89_LParan->Show();
    K89_RParan->Show(); K89_Comma->Show(); K89_Div->Show();
    K89_I->Show(); K89_7->Show(); K89_8->Show();
    K89_9->Show(); K89_Mult->Show(); K89_EE->Show();
    K89_4->Show(); K89_5->Show(); K89_6->Show();
    K89_Minus->Show(); K89_Sto->Show(); K89_1->Show();
    K89_2->Show(); K89_3->Show(); K89_Plus->Show();
    K89_On->Show(); K89_0->Show(); K89_Dot->Show();
    K89_Neg->Show(); K89_Enter->Show();
}

int lcdUpdate;
void __fastcall TEmu92Wnd::FormCreate(TObject *Sender)
{
    if (calc==89)
        Make89();
    else
        Make92();

    lcdUpdate=0;

    Application->OnIdle=TEmu92Wnd::OnIdle;
    Application->OnActivate=TEmu92Wnd::OnActivate;
    Application->OnDeactivate=TEmu92Wnd::OnDeactivate;
    bmp=new Graphics::TBitmap;
    bmp->Width=240;
    bmp->Height=128;
    bmp->PixelFormat=pf32bit;
    for (int i=0;i<128;i++)
        ScreenLine[i]=(char *)bmp->ScanLine[i];

	// Create DirectInput object
    lpDI=NULL; lpKeyboard=NULL;
	if (DirectInputCreate(hInst,DIRECTINPUT_VERSION,&lpDI,NULL)!=DI_OK)
    {
		MessageBox(NULL,"Could not initialize DirectInput","Error",MB_OK);
        return;
    }
	// Create and set up the keyboard device
	if (lpDI->CreateDevice(GUID_SysKeyboard,(LPDIRECTINPUTDEVICE*)&lpKeyboard,
		NULL)!=DI_OK)
    {
		MessageBox(NULL,"Could not access keyboard from DirectInput","Error",MB_OK);
        return;
    }
	lpKeyboard->SetDataFormat(&c_dfDIKeyboard);
	lpKeyboard->SetCooperativeLevel(Handle,DISCL_EXCLUSIVE|
		DISCL_FOREGROUND);
	DIPROPDWORD dpd;
	dpd.diph.dwSize=sizeof(DIPROPDWORD);
	dpd.diph.dwHeaderSize=sizeof(DIPROPHEADER);
	dpd.diph.dwObj=0;
	dpd.diph.dwHow=DIPH_DEVICE;
	dpd.dwData=64;
	lpKeyboard->SetProperty(DIPROP_BUFFERSIZE,&dpd.diph);

    lpKeyboard->Acquire();

    Restricttoactualspeed1->Checked=restrict;
}
//---------------------------------------------------------------------------
extern int contrast,debug,run,debugStartPC;
int showSpeed=0;
void __fastcall TEmu92Wnd::OnIdle(TObject *,bool& done)
{
    done=false;
    if ((!debug)||(run))
    {
        int i;
        for (i=0;i<=31;i++)
        {
            DoInstruction();
            DoInstruction();
            DoInstruction();
            DoInstruction();
            DoInstruction();
            DoInstruction();
            DoInstruction();
            DoInstruction();
            DoInstruction();
            DoInstruction();
        }
        if (restrict)
        {
            static int lastTime=GetTickCount();
            int curTime=GetTickCount();
            int time=curTime-lastTime;
            if (time<20)
                Sleep(20-time);
            lastTime=curTime;
        }
        if (showSpeed)
        {
            static int ticks=0; ticks+=200;
            static int lastTime=GetTickCount();
            int curTime=GetTickCount();
            if ((curTime-lastTime)>=1000)
            {
                char str[64];
                sprintf(str,"%s - %d%% emulation speed",
                    (calc==89)?"Virtual TI-89":(plusMod?
                    "Virtual TI-92+":"Virtual TI-92"),
                    (ticks*100)/(9875*(curTime-lastTime)/1000));
                Caption=str;
                lastTime=curTime;
                ticks=0;
            }
        }
        static int lcdRefresh=0; lcdRefresh++;
        if (lcdRefresh>1)
        {
            lcdRefresh=0;
            LCD->Canvas->Draw(0,0,bmp);
        }
        if (emuLink!=openLink)
        {
            clCloseCom();
            if (emuLink)
            {
                clInitCom(emuLink);
                openLink=emuLink;
            }
        }
        if (openLink)
        {
            int c;
            while (clGetCom(c))
                SendCom(c);
            if (CheckCom())
            {
                int c;
                GetCom(&c);
                clSendCom(c);
            }
        }
        if (lpKeyboard)
        {
            DIDEVICEOBJECTDATA data[64];
            DWORD count=64;
            int err=lpKeyboard->GetDeviceData(sizeof(DIDEVICEOBJECTDATA),
                data,&count,0);
            if (err==DIERR_INPUTLOST)
            {
                lpKeyboard->Acquire();
                return;
            }
            else if (err!=DI_OK)
                return;
            for (i=0;i<count;i++)
            {
                int j;
                if ((data[i].dwOfs==DIK_F10)&&(data[i].dwData&0x80))
                    SendFile1Click(NULL);
                else if ((data[i].dwOfs==DIK_F11)&&(data[i].dwData&0x80))
                    Enterdebugmode1Click(NULL);
                else if ((data[i].dwOfs==DIK_F12)&&(data[i].dwData&0x80))
                    Setemulationmode1Click(NULL);
                else
                {
                    for (j=0;keymap[j]!=-1;j+=2)
                    {
                        if (data[i].dwOfs==keymap[j])
                        {
                            if (data[i].dwData&0x80)
                                kdown(keymap[j+1]);
                            else
                                kup(keymap[j+1]);
                            break;
                        }
                    }
                }
            }
        }
    }
    else
    {
        if (lcdUpdate)
        {
            LCD->Canvas->Draw(0,0,bmp);
            lcdUpdate=0;
        }
    }
}


void __fastcall TEmu92Wnd::LCDPaint(TObject *Sender)
{
    if (!bmp) return;
    LCD->Canvas->Draw(0,0,bmp);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF2MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f2);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF2MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f2);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF1MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f1);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF1MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f1);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF3MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f3);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF3MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f3);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF4MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f4);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF4MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f4);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF5MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f5);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF5MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f5);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF6MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f6);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF6MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f6);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF7MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f7);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF7MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f7);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF8MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f8);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyF8MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f8);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyHandMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_hand);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyHandMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_hand);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyUpMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_up);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyUpMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_up);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyDownMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_down);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyDownMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_down);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyLeftMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_left);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyLeftMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_left);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyRightMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_right);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyRightMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_right);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEnter2MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_enter2);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEnter2MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_enter2);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyQMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_q);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyQMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_q);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyWMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_w);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyWMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_w);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_e);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_e);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyRMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_r);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyRMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_r);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyTMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_t);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyTMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_t);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyYMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_y);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyYMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_y);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyUMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_u);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyUMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_u);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyIMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_i);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyIMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_i);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyOMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_o);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyOMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_o);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyPMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_p);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyPMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_p);    
}
//---------------------------------------------------------------------------


void __fastcall TEmu92Wnd::KeyAMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_a);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyAMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_a);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeySMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_s);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeySMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_s);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyDMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_d);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyDMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_d);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyFMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyFMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyGMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_g);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyGMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_g);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyHMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_h);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyHMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_h);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyJMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_j);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyJMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_j);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyKMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_k);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyKMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_k);    
}
//---------------------------------------------------------------------------


void __fastcall TEmu92Wnd::KeyLMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_l);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyLMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_l);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyShiftMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_shift);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyShiftMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_shift);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyZMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_z);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyZMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_z);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyXMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_x);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyXMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_x);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyCMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_c);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyCMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_c);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyVMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_v);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyVMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_v);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyBMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_b);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyBMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_b);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyNMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_n);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyNMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_n);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyMMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_m);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyMMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_m);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyThetaMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_theta);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyThetaMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_theta);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEnter1MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_enter1);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEnter1MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_enter1);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyDiaMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_dia);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyDiaMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_dia);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key2ndAMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_2nd);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key2ndAMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_2nd);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key2ndBMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_2nd);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key2ndBMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_2nd);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyStoMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_sto);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyStoMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_sto);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeySpaceMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_space);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeySpaceMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_space);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEquMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_equ);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEquMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_equ);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyBSMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_bs);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyBSMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_bs);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEscAMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_esc);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEscAMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_esc);    
}
//---------------------------------------------------------------------------


void __fastcall TEmu92Wnd::KeyModeMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_mode);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyModeMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_mode);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyClearMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_clear);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyClearMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_clear);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyAppsMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_apps);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyAppsMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_apps);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEscBMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_esc);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEscBMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_esc);    
}
//---------------------------------------------------------------------------


void __fastcall TEmu92Wnd::KeyLnMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_ln);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyLnMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_ln);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeySinMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_sin);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeySinMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_sin);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyCosMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_cos);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyCosMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_cos);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyTanMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_tan);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyTanMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_tan);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyPowerMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_power);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyPowerMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_power);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyLParanMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_lparan);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyLParanMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_lparan);
}

void __fastcall TEmu92Wnd::KeyRParanMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_rparan);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyRParanMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_rparan);
}

void __fastcall TEmu92Wnd::KeyCommaMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_comma);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyCommaMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_comma);
}

void __fastcall TEmu92Wnd::KeyDivMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_div);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyDivMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_div);
}

void __fastcall TEmu92Wnd::KeyMultMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_mult);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyMultMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_mult);
}

void __fastcall TEmu92Wnd::KeyMinusMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_minus);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyMinusMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_minus);
}

void __fastcall TEmu92Wnd::KeyPlusMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_plus);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyPlusMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_plus);
}

void __fastcall TEmu92Wnd::Key0MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_0);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key0MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_0);
}

void __fastcall TEmu92Wnd::Key1MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_1);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key1MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_1);
}

void __fastcall TEmu92Wnd::Key2MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_2);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key2MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_2);
}

void __fastcall TEmu92Wnd::Key3MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_3);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key3MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_3);
}

void __fastcall TEmu92Wnd::Key4MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_4);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key4MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_4);
}

void __fastcall TEmu92Wnd::Key5MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_5);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key5MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_5);
}

void __fastcall TEmu92Wnd::Key6MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_6);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key6MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_6);
}

void __fastcall TEmu92Wnd::Key7MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_7);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key7MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_7);
}

void __fastcall TEmu92Wnd::Key8MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_8);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key8MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_8);
}

void __fastcall TEmu92Wnd::Key9MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_9);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Key9MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_9);
}

void __fastcall TEmu92Wnd::KeyDotMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_dot);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyDotMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_dot);
}

void __fastcall TEmu92Wnd::KeyNegMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_neg);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyNegMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_neg);
}

void __fastcall TEmu92Wnd::KeyEnter3MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_enter1);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyEnter3MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_enter1);
}


void __fastcall TEmu92Wnd::LCDMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    TPoint p;
    if (Button==mbRight)
    {
        p.x=X; p.y=Y; p=LCD->ClientToScreen(p);
        PopupMenu1->Popup(p.x,p.y);
    }
}
//---------------------------------------------------------------------------

extern int TransferFile(char *fn);
void __fastcall TEmu92Wnd::SendFile1Click(TObject *Sender)
{
    TOpenDialog *dlog=new TOpenDialog(this);
    dlog->DefaultExt=(calc==89)?"89z":(plusMod?"9xz":"92p");
    dlog->FileName=(calc==89)?"*.89?":(plusMod?"*.9x?":"*.92?");
    dlog->Options.Clear();
    dlog->Options << ofFileMustExist << ofPathMustExist << ofAllowMultiSelect;
    dlog->Filter="TI-89 files (*.89?)|*.89?|TI-92 files (*.92?)|*.92?|TI-92 Plus files (*.9x?)|*.9x?";
    dlog->FilterIndex=(calc==89)?1:(plusMod?3:2);
    if (useInitPath)
    {
        useInitPath=0;
        dlog->InitialDir=initPath;
    }
    if (dlog->Execute())
    {
        for (int i=0;i<dlog->Files->Count;)
        {
            if (!TransferFile(dlog->Files->Strings[i].c_str()))
            {
                i++;
                continue;
            }
            if (MessageBox(Handle,"Transmission error","Error",
                MB_RETRYCANCEL|MB_ICONEXCLAMATION)==IDCANCEL)
                break;
        }
    }
    delete dlog;
}
//---------------------------------------------------------------------------

extern void ResetEmu();
void __fastcall TEmu92Wnd::Reset1Click(TObject *Sender)
{
    ResetEmu();
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::FormDestroy(TObject *Sender)
{
    TRegistry *reg=new TRegistry;
    reg->RootKey=HKEY_LOCAL_MACHINE;
    reg->OpenKey("\\Software\\PlusShell",true);
    reg->WriteString("LastEmuROM",romImage[currentROM].name);
    reg->WriteInteger("RestrictSpeed",restrict);
    reg->OpenKey("EmuKeyMap",true);
    for (int i=k_q;i<=k_m;i++)
    {
        int j;
        for (j=0;keymap[j]!=-1;j+=2)
        {
            if (keymap[j+1]==i)
            {
                char str[4];
                sprintf(str,"%s",alphakeystr[i-k_q]);
                reg->WriteInteger(str,keymap[j]);
                break;
            }
        }
    }
    reg->CloseKey();
    if (saveStateOnExit)
    {
        chdir(initPath);
        SaveState(NULL);
    }
    if (lpKeyboard) lpKeyboard->Unacquire();
    if (lpKeyboard) lpKeyboard->Release();
    lpKeyboard=NULL;
    if (lpDI) lpDI->Release();
    if (openLink) clCloseCom();
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyOnMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_on);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::KeyOnMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_on);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::COM1Click(TObject *Sender)
{
    if (emuLink==1)
    {
        emuLink=0;
        COM1->Checked=false;
    }
    else
    {
        emuLink=1;
        COM1->Checked=true;
        COM2->Checked=false;
        COM3->Checked=false;
        COM4->Checked=false;
    }
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::COM2Click(TObject *Sender)
{
    if (emuLink==2)
    {
        emuLink=0;
        COM2->Checked=false;
    }
    else
    {
        emuLink=2;
        COM1->Checked=false;
        COM2->Checked=true;
        COM3->Checked=false;
        COM4->Checked=false;
    }
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::COM3Click(TObject *Sender)
{
    if (emuLink==3)
    {
        emuLink=0;
        COM3->Checked=false;
    }
    else
    {
        emuLink=3;
        COM1->Checked=false;
        COM2->Checked=false;
        COM3->Checked=true;
        COM4->Checked=false;
    }
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::COM4Click(TObject *Sender)
{
    if (emuLink==4)
    {
        emuLink=0;
        COM4->Checked=false;
    }
    else
    {
        emuLink=4;
        COM1->Checked=false;
        COM2->Checked=false;
        COM3->Checked=false;
        COM4->Checked=true;
    }
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::FormActivate(TObject *Sender)
{
    if (lpKeyboard) lpKeyboard->Acquire();    
}
//---------------------------------------------------------------------------

bool unacquire=true;
void __fastcall TEmu92Wnd::FormDeactivate(TObject *Sender)
{
    if (!unacquire) return;
    if (lpKeyboard) lpKeyboard->Unacquire();
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::About1Click(TObject *Sender)
{
    AboutDlg->ShowModal();    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Enterdebugmode1Click(TObject *Sender)
{
    debugStartPC=regs.pc;
    DebugWnd->Show();
    DebugWnd->Update();
    debug=1; run=0;
}
//---------------------------------------------------------------------------

#define MODE_TI89 0
#define MODE_TI92 1
#define MODE_TI92PLUS 2
#define MODE_TI92PLUSINT 3

extern void CloseEmu();
extern int InitEmu(int mode);
extern int currentROM;
void __fastcall TEmu92Wnd::Setemulationmode1Click(TObject *Sender)
{
    run=0;
    EmuModeDlg->Calc->ItemIndex=currentROM;
    if (EmuModeDlg->ShowModal()==1)
    {
        if (EmuModeDlg->Calc->ItemIndex!=-1)
        {
            chdir(initPath);
            SaveState(NULL);
            CloseEmu();
            InitEmu(EmuModeDlg->Calc->ItemIndex);
            if (calc==89)
                Make89();
            else
                Make92();
            LoadState(NULL);
            AboutDlg->UpdateText();
            useInitPath=1;
        }
    }
    run=1;
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_F1MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f1);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_F1MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f1);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_F2MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f2);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_F2MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f2);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_F3MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f3);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_F3MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f3);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_F4MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f4);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_F4MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f4);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_F5MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f5);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_F5MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f5);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_2ndMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_2nd);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_2ndMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_2nd);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_ShiftMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_shift);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_ShiftMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_shift);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_EscMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_esc);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_EscMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_esc);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_LeftMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_left);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_LeftMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_left);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_UpMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_up);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_UpMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_up);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_RightMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_right);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_RightMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_right);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_DownMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_down);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_DownMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_down);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_DiaMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_dia);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_DiaMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_dia);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_AlphaMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_hand);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_AlphaMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_hand);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_AppsMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_apps);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_AppsMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_apps);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_HomeMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_cos);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_HomeMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_cos);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_ModeMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_mode);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_ModeMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_mode);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_CatalogMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_f6);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_CatalogMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_f6);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_BSMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_bs);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_BSMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_bs);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_ClearMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_clear);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_ClearMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_clear);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_XMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_x);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_XMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_x);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_YMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_y);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_YMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_y);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_ZMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_z);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_ZMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_z);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_TMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_t);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_TMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_t);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_PowerMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_power);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_PowerMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_power);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_EquMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_equ);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_EquMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_equ);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_LParanMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_lparan);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_LParanMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_lparan);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_RParanMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_rparan);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_RParanMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_rparan);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_CommaMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_comma);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_CommaMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_comma);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_DivMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_div);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_DivMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_div);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_IMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_ln);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_IMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_ln);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_7MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_7);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_7MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_7);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_8MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_8);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_8MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_8);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_9MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_9);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_9MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_9);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_MultMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_mult);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_MultMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_mult);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_EEMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_sin);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_EEMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_sin);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_4MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_4);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_4MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_4);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_5MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_5);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_5MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_5);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_6MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_6);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_6MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_6);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_MinusMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_minus);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_MinusMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_minus);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_StoMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_sto);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_StoMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_sto);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_1MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_1);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_1MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_1);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_2MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_2);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_2MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_2);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_3MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_3);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_3MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_3);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_PlusMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_plus);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_PlusMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_plus);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_OnMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_on);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_OnMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_on);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_0MouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_0);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_0MouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_0);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_DotMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_dot);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_DotMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_dot);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_NegMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_neg);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_NegMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_neg);    
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_EnterMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kdown(k_enter1);
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::K89_EnterMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    kup(k_enter1);
}
//---------------------------------------------------------------------------

extern int debugProgBreak,debugProgBreakOfs;
void __fastcall TEmu92Wnd::Setprogramentrybreakpoint1Click(TObject *Sender)
{
    ProgramEntryDlg->UpdateTree();
    if (ProgramEntryDlg->ShowModal()==1)
    {
        debugProgBreak=((((int)ProgramEntryDlg->Programs->
            Selected->Data)&0xffff)<<2)+HANDLE_TABLE;
        debugProgBreakOfs=(((int)ProgramEntryDlg->Programs->
            Selected->Data)>>16)&0xffff;
    }
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Savestateimage1Click(TObject *Sender)
{
    TSaveDialog *dlog=new TSaveDialog(this);
    dlog->DefaultExt="sav";
    dlog->FileName=(calc==89)?"ti89.sav":(plusMod?"ti92plus.sav":"ti92.sav");
    dlog->Options.Clear();
    dlog->Filter="State files (*.sav)|*.sav";
    dlog->FilterIndex=1;
    if (useInitPath)
    {
        useInitPath=0;
        dlog->InitialDir=initPath;
    }
    if (dlog->Execute())
        SaveState(dlog->FileName.c_str());
    delete dlog;
}
//---------------------------------------------------------------------------

extern int calcChanged;
void __fastcall TEmu92Wnd::Loadstateimage1Click(TObject *Sender)
{
    TOpenDialog *dlog=new TOpenDialog(this);
    dlog->DefaultExt="sav";
    dlog->FileName="*.sav";
    dlog->Options.Clear();
    dlog->Options << ofFileMustExist << ofPathMustExist;
    dlog->Filter="State files (*.sav)|*.sav";
    dlog->FilterIndex=1;
    if (useInitPath)
    {
        useInitPath=0;
        dlog->InitialDir=initPath;
    }
    if (dlog->Execute())
    {
        char old[256];
        getcwd(old,256);
        chdir(initPath);
        SaveState(NULL);
        calcChanged=0;
        LoadState(dlog->FileName.c_str());
        if (calcChanged)
        {
            if (calc==89) Make89();
            else Make92();
            AboutDlg->UpdateText();
        }
        chdir(old);
    }
    delete dlog;
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Exitwithoutsavingstate1Click(TObject *Sender)
{
    chdir(initPath);
    LoadState(NULL);
    useInitPath=1;
}
//---------------------------------------------------------------------------

extern int oldEmu;
void __fastcall TEmu92Wnd::UseNewClick(TObject *Sender)
{
    MakeSR();
    oldEmu=!oldEmu;
//    UseNew->Checked=!oldEmu;
    MakeFromSR();
    MakeSR();
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::ShowSpeed1Click(TObject *Sender)
{
    showSpeed=!showSpeed;
//    ShowSpeed1->Checked=showSpeed;
    Caption=(calc==89)?"Virtual TI-89":(plusMod?
        "Virtual TI-92+":"Virtual TI-92");
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
    Key=0;
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::FormKeyUp(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
    Key=0;
}
//---------------------------------------------------------------------------

extern int DumpROM();
void __fastcall TEmu92Wnd::DumpROM1Click(TObject *Sender)
{
    Hide();
    DebugWnd->Hide();
    DumpROM();
    Show();
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Restricttoactualspeed1Click(TObject *Sender)
{
    restrict=!restrict;
    Restricttoactualspeed1->Checked=restrict;
}
//---------------------------------------------------------------------------

void __fastcall TEmu92Wnd::Redefinekeys1Click(TObject *Sender)
{
    if (!lpKeyboard) return;
    unacquire=false;
    RedefineWnd->Show();
    RedefineWnd->Update();
    int held=0;
    RedefineWnd->Key->Caption="Press new key for \"Q\"";
    for (int n=k_q;n<=k_m;)
    {
        DIDEVICEOBJECTDATA data[64];
        DWORD count=64;
        int err=lpKeyboard->GetDeviceData(sizeof(DIDEVICEOBJECTDATA),
            data,&count,0);
        if (err==DIERR_INPUTLOST)
        {
            lpKeyboard->Acquire();
            continue;
        }
        else if (err!=DI_OK)
        {
            RedefineWnd->Hide();
            unacquire=true;
            return;
        }
        for (int i=0;i<count;i++)
        {
            int j;
            if (data[i].dwData&0x80)
            {
                if (held) continue;
                for (j=0;keymap[j]!=-1;j+=2)
                {
                    if (n==keymap[j+1])
                    {
                        held=1;
                        keymap[j]=data[i].dwOfs;
                        break;
                    }
                }
            }
            else if (held)
            {
                held=0;
                n++;
                if (n>k_m) break;
                RedefineWnd->Key->Caption=AnsiString("Press new key for \"")+
                    AnsiString(alphakeystr[n-k_q])+AnsiString("\"");
                RedefineWnd->Update();
            }
        }
    }
    RedefineWnd->Hide();
    unacquire=true;
}
//---------------------------------------------------------------------------

